#include "colourpad.h"
#include "canvas.h"
#include "window.h"

extern "C"
{
#  include "support.h"
}

namespace sudoku
{

/*********************************************************************
 */
void
on_colour_toggled(GtkToggleToolButton *button, gpointer user_data)
{
  if ( gtk_toggle_tool_button_get_active(button) )
  {
    ColourPad::Key *key = static_cast<ColourPad::Key*>(user_data);
    key->getPad()->setSelection( key->getData() );
  }
}


void
on_colour_realize(GtkWidget*, gpointer user_data)
{
  ColourPad::Key *key = static_cast<ColourPad::Key*>(user_data);

  const int colour = key->getData();
  if ( (colour == -1) ||
       gdk_colormap_alloc_color(gdk_colormap_get_system(),
                                &ColourPad::COLOUR[colour],
                                FALSE,
                                TRUE) )
  {
    key->loadIcon();
  }
}


/*********************************************************************
 */
const guchar ColourPad::Key::colourXbm[(ICON_SIZE * ICON_SIZE) / CHAR_BIT] = {
       0xFF, 0xFF,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0xFF, 0xFF
};


const guchar ColourPad::Key::numberXbm[(ICON_SIZE * ICON_SIZE) / CHAR_BIT] = {
       0xFF, 0xFF,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x83,
       0xC1, 0x83,
       0x01, 0x83,
       0x01, 0x83,
       0x01, 0x83,
       0x01, 0x83,
       0x01, 0x83,
       0x01, 0x83,
       0xC1, 0x8F,
       0x01, 0x80,
       0x01, 0x80,
       0x01, 0x80,
       0xFF, 0xFF
};


#define _RGB_(r, g, b)  (r)*256, (g)*256, (b)*256

GdkColor ColourPad::COLOUR[MAX_COLOURS] = {
                       { 0, _RGB_(255, 255, 255) },  // White
                       { 0, _RGB_(142, 255, 135) },  // Dark green
                       { 0, _RGB_(210, 255, 108) },  // Light green
                       { 0, _RGB_(252, 252,   7) },  // Yellow
                       { 0, _RGB_(249, 227, 106) },  // Orange
                       { 0, _RGB_(246, 188, 238) },  // Pink
                       { 0, _RGB_(255, 181, 151) },  // Red
                       { 0, _RGB_(198, 218, 255) },  // Lilac
                       { 0, _RGB_(170, 255, 247) }   // Blue
};


/*********************************************************************
 */
ColourPad::Key::Key(ColourPad *pad,
                    GSList *group,
                    int data)
  : m_key( gtk_radio_tool_button_new(group) ) 
  , m_icon(NULL)
  , m_toggleHandler(0)
  , m_realizeHandler(0)
  , m_data(data)
  , m_pad(pad)
{
  gtk_widget_set_size_request(GTK_WIDGET(m_key), BUTTON_SIZE, BUTTON_SIZE);
  gtk_widget_show( GTK_WIDGET(m_key) );
}


/*********************************************************************
 */
void
ColourPad::Key::loadIcon()
{
  const guchar *xbmData = colourXbm;
  const GdkColor *colour;

  if (m_data == -1)
  {
    xbmData = numberXbm;
    colour = ColourPad::getColour(0);
  }
  else
  {
    colour = ColourPad::getColour(m_data);
  }

  /*
   * We must have allocated the colours before creating the pixmaps.
   */
  GdkPixmap *pixmap = gdk_pixmap_create_from_data(
                          GTK_WIDGET(m_key)->window,
                          reinterpret_cast<const gchar*>(xbmData),
                          ICON_SIZE, ICON_SIZE,
                          -1,
                          &GTK_WIDGET(m_key)->style->black,
                          colour);

  m_icon = gtk_image_new_from_pixmap(pixmap, NULL);
  gtk_widget_show(m_icon);

  gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON(m_key), m_icon );

  if (pixmap != NULL)
    g_object_unref(pixmap);
}


/*********************************************************************
 */
void
ColourPad::Key::setSignals()
{
  m_toggleHandler = g_signal_connect(static_cast<gpointer>(m_key),
                                     "toggled",
                                     G_CALLBACK(on_colour_toggled),
                                     this);
  m_realizeHandler = g_signal_connect(static_cast<gpointer>(m_key),
                                     "realize",
                                     G_CALLBACK(on_colour_realize),
                                     this);
}


/*********************************************************************
 */
GSList*
ColourPad::createKey(int keyValue, GSList *group)
{
  m_keys.push_back( Key(this, group, keyValue) );

  Key &key = m_keys.back();
  key.setSignals();

  GtkToolItem *gkey = key.getKey();
  group = gtk_radio_tool_button_get_group( GTK_RADIO_TOOL_BUTTON(gkey) );
  gtk_toolbar_insert(GTK_TOOLBAR(getGtkWidget()), gkey, -1);

  return group;
}


/*********************************************************************
 */
ColourPad::~ColourPad()
{
  /*
   * I don't this I need to deallocate the colours
   * after all - the colour array within the colour
   * map is NULL.
   *
   * deallocateColours();
   */
}


ColourPad::ColourPad(Window *w)
  : Component( w, gtk_toolbar_new() )
  , m_selection(-1)
{
  GtkWidget *pad = lookup_widget(w->getGtkWidget(), "colour_align");
  gtk_container_add(GTK_CONTAINER(pad), getGtkWidget());
  gtk_toolbar_set_style(GTK_TOOLBAR(getGtkWidget()), GTK_TOOLBAR_ICONS);

  GSList *group = NULL;

  for (unsigned i = 0; i < ARRAY_SIZE(COLOUR); ++i)
  {
    group = createKey(i, group);
  } /* for */

  GtkToolItem* separator = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(getGtkWidget()), separator, -1);
  gtk_widget_show( GTK_WIDGET(separator) );

  group = createKey(-1, group);

  disable();
  show();
}


/*********************************************************************
 */
void
ColourPad::disable()
{
  gtk_toggle_tool_button_set_active(
                        GTK_TOGGLE_TOOL_BUTTON(m_keys.back().getKey()), TRUE);
  disableWidget(getGtkWidget());
}


/*********************************************************************
 *
void
ColourPad::deallocateColours()
{
  gdk_colormap_free_colors( gdk_colormap_get_system(),
                            COLOUR,
                            ARRAY_SIZE(COLOUR) );
}
 */

}

